#!/bin/bash
# Using slurm's command to get the processes
# Binding pthreads within an MPI task to the cores (within the taskset)
# Binding pthreads to cores when not using MPI
# Author: Manu Shantharam
#

function debug {
  if [ "$AFFINITY_DEBUG" == "1" ]; then
    echo -e "AFF-DEBUG: $@"
  fi
}

function infop {
  if [ "$AFFINITY_INFO" == "1" ]; then
    echo -e "AFF-INFO: $@"
  fi
}


# if the partition is compute, just exit
#if [[ $SLURM_JOB_PARTITION == "compute" ]]; then
#   infop "This script doesn't support compute partition at present"
#   exit 0
#fi

# get the binary name and number of MPI tasks
nametmp=$1
np=$SLURM_TASKS_PER_NODE
nt=$SLURM_CPUS_PER_TASK

debug "Name of the binary: "$nametmp
debug "Number of MPI tasks: "$np
debug "Number of threads ${nt}"

# poll till all processes are running
# change polling frequency after 5 seconds as we don't want to overload slurm
poll=120
while [[ $poll -gt 0 ]]; do
  ((poll--))
  if [[ $poll -gt 115 ]]; then # first 5 seconds sleep for 1 second
     debug "Sleeping for 1s"
     sleep 1
  else                         # after that sleep for 20 seconds  
     debug "Sleeping for 20s"
     sleep 20
  fi
  p=(`scontrol listpids ${SLURM_JOBID} | awk '{if ($1 != "PID") print $1}'`)
  debug "PIDs: ${p[@]}"
  idx=0
  ptmp=()
  found=0
  for i in ${p[@]};
  do
     if [[ $i !=  "-1" ]]; then
         ps -p $i ww -o cmd | awk '{print $1}' | grep -q $nametmp
         if [[ $? -eq 0 ]]; then
	    debug "COMMAND (PID $i): " `ps -p $i ww -o cmd`
	    ptmp[$idx]=$i
	    ((idx++))
	    found=1
         fi
     fi
  done
  if [[ $found -eq 1  && ${#ptmp[@]} -eq $np ]]; then
     infop "Found all the process(es): ${ptmp[@]}"
     break
  fi
done

# poll for all active threads with all processes
# we assume all processes have active threads
p=( "${ptmp[@]}" )


# store taskset for all process
init_ts=()
for i in ${p[@]};
do
  # get the taskset of the main MPI task
  ts=`taskset -cp $i |  awk '{print $NF}'`
  ts_fmt=`echo $ts |sed -re 's:([0-9]+)[-]([0-9]+):{\1..\2}:g'`
  echo ${ts_fmt} | grep -q ","
  if [[ $? != 0 ]]; then
     ts_exp=(`eval echo $ts_fmt`)
  else
     ts_exp=(`eval echo "{"$ts_fmt"}"`)
  fi
  init_ts=(`echo ${init_ts[@]}; echo ${ts_exp[@]} | sed "s| |-|g"`)
done

infop "Initial taskset: ${init_ts[@]}"

idx=0
cur_at_list=()
prev_at_list=()
while [[  1 ]]; do

  ((idx++))
  infop "Polling for active threads: iteration number $idx"

  poll=200
  foundallats=0
  while [[ $poll -gt 0 ]]; do
    sleep 1  
    ((poll--))
    foundp=0
    cur_at_list=()
    for i in ${p[@]};
    do
      at=(`ps -T -o pid,tid,state -p $i |  awk '{if ($3 != "S") print $2}'`)
      debug "Number of active threads ($i): " ${#at[@]}
      if [[ ${#at[@]} -eq  $nt ]]; then
	 cur_at_list=(`echo ${cur_at_list[@]}; echo ${at[@]}`)     
         ((foundp++))
      fi
    done
    if [[ $foundp -eq ${#p[@]} ]]; then
#       infop "All processes have active threads now. Will starting binding the threads."
       foundallats=1
       break
    fi
  done

#  if [[ "${prev_at_list[*]}" == "${cur_at_list[*]}" ]]; then
#     continue    # same set of active threads, so no binding required
#  else
#     infop "=============================Previous active thread list: ${prev_at_list[@]}"
#     infop "=============================Current active thread list: ${cur_at_list[@]}" 
#     infop "All processes have new active threads now. Will starting binding the threads."
#     prev_at_list=(`echo ${cur_at_list[@]}`)
#     cur_at_list=()
#  fi 
#  

  # try to bind after every 20 + (1 to 200) seconds
  sleep 20
  
  if [[ $foundallats -eq 0 ]]; then
     infop "At least one of the processes does not seem to have active threads. Anyway trying to bind...."
  fi 
  
  idx1=0
  for i in ${p[@]}; 
  do
    ts_exp=(`echo ${init_ts[$idx1]}| sed "s|-| |g"`)
    ((idx1++))
    
    # get all the active threads for this task
    # assumption: number of threads <= number of cores in the taskset
    at=(`ps -T -o pid,tid,state -p $i |  awk '{if ($3 != "S") print $2}'`)
  
    debug "Number of active threads ($i -- taskset:${ts_exp[@]}): " ${#at[@]}
    if [[ ${#at[@]} -lt 2 ]]; then
       infop "No threads to bind"
       continue
    fi
  
    # bind threads to cores from the taskset (ts_exp)
    for ((j=0; j<${#at[@]}; j=j+1 ))
    do
  #     echo $ts_exp[$j]"---------" ${at[$j]}       
       out=`taskset -cp ${ts_exp[$j]} ${at[$j]}`
       if [[ $? != 0 ]]; then
         infop "Could not bind: " $out
         echo "Could not bind: " $out
          continue
       fi
       debug "Binding process ${at[$j]} to core ${ts_exp[$j]}"
    done
  done
done
